/*-
 * Copyright (c) 2000-2001 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: cmd.c,v 1.2.2.4 2001/03/03 23:02:33 simon Exp $";
#endif /* !lint */


#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/param.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include "cmd.h"

#include "debug.h"
#include "ipa.h"


/*
 * Send message to syslog about some problems with command.
 */
static void
exec_failed_log(u_int no, const char *message, va_list ap)
{
	vsyslog(LOG_ERR, message, ap);
	syslog(LOG_ERR, "  -> command in exec #%u failed", no + 1);
}

/*
 * Run command: pass command string to system() function.
 */
static int
exec_cmd(struct cmd *cp, u_int no, const char *message, va_list ap)
{
	int	status;

	if (debug_exec > 0) {
		uid_t		uid;
		gid_t		gid;
		struct passwd	*pwd;
		struct group	*grp;

		uid = getuid();
		gid = getgid();
		pwd = getpwuid(uid);
		grp = getgrgid(gid);
		if (debug_exec == 1)
			syslog(LOG_INFO, "exec_cmd: exec #%u by %s:%s, UID %d GID %d",
			    no + 1, pwd ? pwd->pw_name : "(UNKNOWN USER)",
			    grp ? grp->gr_name : "(UNKNOWN GROUP)", uid, gid);
		else
			syslog(LOG_INFO, "exec_cmd: exec #%u (%s) by %s:%s, UID %d GID %d",
			    no + 1, cp->str, pwd ? pwd->pw_name : "(UNKNOWN USER)",
			    grp ? grp->gr_name : "(UNKNOWN GROUP)", uid, gid);
	}
	if ( (status = system(cp->str)) < 0)
		syslog(LOG_ERR, "exec_cmd: system: %m");
	else if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) == 127) {
			syslog(LOG_ERR, "exec_cmd: system: exit status = 127. Was execution of the shell failed?");
			status = -1;
		}
	} else if (WIFSIGNALED(status)) {
		syslog(LOG_ERR, "exec_cmd: system: child abnormal termination, signal %d '%s'%s",
		    WTERMSIG(status), sys_signame[WTERMSIG(status)],
		    WCOREDUMP(status) ? " (core file generated)" : "");
		status = -1;
	} else if (WIFSTOPPED(status)) {
		syslog(LOG_ERR, "exec_cmd: system: child stopped, signal %d '%s'",
		    WSTOPSIG(status), sys_signame[WSTOPSIG(status)]);
		status = -1;
	}
	if (status < 0) {
		exec_failed_log(no, message, ap);
		return -1;
	}
	return 0;
}

/*
 * Fork() child, change UID, GID and supplementary GIDs if needed
 * and run command.
 */
static int
exec_cmd_perm(struct cmd *cp, u_int no, const char *message, va_list ap)
{
	int		fd;
	pid_t		childpid;

	if (cp->uid_is_set == 0) {
		/* Need to change only group. Don't need to fork() extra child */
		if (setgid(cp->gid) < 0) {
			syslog(LOG_ERR, "exec_cmd_perm: setgid(%lu): %m", cp->gid);
			exec_failed_log(no, message, ap);
			return -1;
		}
		exec_cmd(cp, no, message, ap);
		if (setgid(mygid) < 0) {
			syslog(LOG_ERR, "can't change GID back, setgid(%u): %m", mygid);
			return -1;
		}
		return 0;
	}

	if ( (childpid = fork()) < 0) {
		syslog(LOG_ERR, "exec_cmd_perm: fork: %m");
		exec_failed_log(no, message, ap);
		return -1;
	}

	if (childpid == 0) {
		/* child */
		struct rlimit	flim;

		getrlimit(RLIMIT_NOFILE, &flim);

		closelog();
		for (fd = 0; fd < flim.rlim_max; ++fd)
			if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO)
				close(fd);
		errno = 0;

		openlog(logfacil, logopt, LOG_USER);
		if (cp->suppl_gid != NULL) {
			/* exec(user:group) or exec(user) */
			if (setgroups(cp->nsuppl_gid, cp->suppl_gid) < 0) {
				syslog(LOG_ERR, "exec_cmd_perm: setgroups: %m");
				exec_failed_log(no, message, ap);
				closelog();
				exit(0);
			}
		}
		if (cp->gid_is_set) {
			if (setgid(cp->gid) < 0) {
				syslog(LOG_ERR, "exec_cmd_perm: setgid(%lu): %m", cp->gid);
				exec_failed_log(no, message, ap);
				closelog();
				exit(0);
			}
		}
		if (setuid(cp->uid) < 0) {
			syslog(LOG_ERR, "exec_cmd_perm: setuid(%lu): %m", cp->uid);
			exec_failed_log(no, message, ap);
			closelog();
			exit(0);
		}
		if (exec_cmd(cp, no, message, ap) < 0 || debug_exec > 0)
			closelog();
		exit(0);
	} else {
		/* parent */
		int	status;

		while (waitpid(childpid, &status, WUNTRACED) < 0)
			if (errno != EINTR) {
				syslog(LOG_ERR, "exec_cmd_perm: waitpid(PID %u): %m", childpid);
				return -1;
			}

		if (WIFSIGNALED(status)) {
			syslog(LOG_ERR, "exec_cmd_perm: child abnormal termination, signal %d '%s'%s",
			    WTERMSIG(status), sys_signame[WTERMSIG(status)],
			    WCOREDUMP(status) ? " (core file generated)" : "");
			return -1;
		} else if (WIFSTOPPED(status)) {
			syslog(LOG_ERR, "exec_cmd_perm: child stopped, signal %d '%s'",
			    WSTOPSIG(status), sys_signame[WSTOPSIG(status)]);
			return -1;
		}
		return 0; 
	}
	/* NOTREACHED */
}

/*
 * Run command list in background and wait for their execution.
 */
int
exec_cmd_list(struct cmd *cp, u_int n, const char *message, ...)
{
	pid_t		childpid;

	if ( (childpid = fork()) < 0) {
		syslog(LOG_ERR, "exec_cmd_list: fork: %m");
		return -1;
	}

	if (childpid == 0) {
		/* child */
		int		fd, fl = 0, ret = 0;
		u_int		i;
		va_list		ap;
		struct rlimit	flim;

		getrlimit(RLIMIT_NOFILE, &flim);

		closelog();
		for (fd = 0; fd < flim.rlim_max; ++fd)
			if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO)
				close(fd);
		errno = 0;

		openlog(logfacil, logopt, LOG_USER);

		va_start(ap, message);
		for (i = 0; i < n; ++cp, ++i) {
			if (cp->uid_is_set || cp->gid_is_set)
				ret = exec_cmd_perm(cp, i, message, ap);
			else {
				ret = exec_cmd(cp, i, message, ap);
				fl = 1;
			}
		}
		va_end(ap);
		if (ret < 0 || (fl == 1 && debug_exec > 0))
			closelog();
		exit(0);
	} else {
		/* parent */
		int	status;

		while (waitpid(childpid, &status, WUNTRACED) < 0)
			if (errno != EINTR) {
				syslog(LOG_ERR, "exec_cmd_list: waitpid(PID %u): %m", childpid);
				return 0;
			}

		if (WIFSIGNALED(status))
			syslog(LOG_ERR, "exec_cmd_list: child abnormal termination, signal %d '%s'%s",
			    WTERMSIG(status), sys_signame[WTERMSIG(status)],
			    WCOREDUMP(status) ? " (core file generated)" : "");
		else if (WIFSTOPPED(status))
			syslog(LOG_ERR, "exec_cmd_list: child stopped, signal %d '%s'",
			    WSTOPSIG(status), sys_signame[WSTOPSIG(status)]);
		return 0;
	}
	/* NOTREACHED */
}

/*
 * Run command list in background.
 */
pid_t
exec_cmd_list_bg(struct cmd *cp, u_int n, const char *message, ...)
{
	u_int		i;
	int		fd, ret = 0, fl = 0;
	pid_t		childpid;
	va_list		ap;
	struct rlimit	flim;

	if ( (childpid = fork()) < 0) {
		syslog(LOG_ERR, "fork: %m");
		return -1;
	}
	if (childpid != 0)
		/* parent */
		return childpid;

	/* child */
	getrlimit(RLIMIT_NOFILE, &flim);
	closelog();
	for (fd = 0; fd < flim.rlim_max; ++fd)
		if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO)
			close(fd);
	errno = 0;

	openlog(logfacil, logopt, LOG_USER);
	va_start(ap, message);
	for (i = 0; i < n; ++cp, ++i) {
		if (cp->uid_is_set || cp->gid_is_set)
			ret = exec_cmd_perm(cp, i, message, ap);
		else {
			ret = exec_cmd(cp, i, message, ap);
			fl = 1;
		}
	}
	va_end(ap);
	if (ret < 0 || (fl == 1 && debug_exec > 0))
		closelog();
	exit(0);
	/* NOTREACHED */
}

/*
 * Release memory used by command list.
 */
void
free_cmd_list(struct cmd *cmd_head, u_int n)
{
	struct cmd	*cmdp;
	u_int		i;

	if (cmd_head) {
		for (cmdp = cmd_head, i = 0; i < n; ++cmdp, ++i) {
			free(cmdp->str);
			free(cmdp->suppl_gid);
		}
		free(cmd_head);
	}
}

/*
 * Release memory used by commands structure.
 */
void
free_commands(struct commands *commandsp)
{
	free_cmd_list(commandsp->cmd, commandsp->ncmd);
	free_cmd_list(commandsp->cmd_if_limit, commandsp->ncmd_if_limit);
	free_cmd_list(commandsp->cmd_if_nolimit, commandsp->ncmd_if_nolimit);
}
